#!/usr/bin/perl

#------------------------------------------------------------------------------
#
#  bluez client for the Bluetooth Location Tracker Project
#
#  invoke with:
#  ./bluez-client.pl <username> <password> <server address> 4444 <latitude> <longitude> [has_rssi]
#
#  Note: In addition to the provided tools hcisniff, inquiry and name_req,
#  this program needs access to the hciconfig tool provided by the bluez
#  packages.
#
#------------------------------------------------------------------------------

use strict;
use Socket;
use POSIX;
use IO::Handle;
use IO::Socket::SSL;
my ($user, $pass, $remote, $port, $iaddr, $paddr, $proto, $line, $has_rssi, $lat, $long);
my @inq_results;
my $ind_msgnr = 0;
my $debug = 0;

#------------------------------------------------------------------------------
#   debug print routine
#------------------------------------------------------------------------------
sub dprint {
    if ($debug != 0) { print $_[0]; }
}

#------------------------------------------------------------------------------
#   evalnumber returns an (decimal) integer of an deciamal or hex string representation
#------------------------------------------------------------------------------
sub evalnumber {
    (my $param = $_[0]) =~ tr/ //d;
    if ($param =~ /x/i) { return oct($param); }
    return int($param);
}


#------------------------------------------------------------------------------
#   filter_inq_results takes a list of inquiry_result_events it got from
#   hcisniff and eliminates duplicate addresses.
#   If there is RSSI information in the inquiry_result_events, an average of
#   all RSSI values corresponding to a single BD_ADDR is built and provided in the
#   final list of filtered inquiry responses.
#   The list of filteres inquiry responses is preformatted for the BLT protocol.
#------------------------------------------------------------------------------
sub filter_inq_results {
    my @filtered_results = ();
    my $known;
    my %rssi_val;
    my %rssi_cnt;

    foreach my $res (@_) {
        (my $addr = $res) =~ s/.*bdaddr ([^,]*).*/$1/;
        $addr = uc($addr);
        (my $rssi = $res) =~ s/.*rssi: ([^,]*).*/$1/;
        $rssi = evalnumber($rssi);
        if ($rssi != 0) {
            $rssi = -65536 + $rssi;
            dprint "got rssi $rssi for $addr\n";
            dprint "val:$rssi_val{$addr} cnt:$rssi_cnt{$addr}\n";
            $rssi_cnt{$addr} += 1;
            $rssi_val{$addr} += $rssi;
        }
        $known = 0;
        foreach my $fres (@filtered_results) {
            (my $faddr = $fres) =~ s/.*inquiry_result_event,[^,]*, ([^,]*),.*/$1/;
            $faddr = uc($faddr);

            if ($addr eq $faddr) { $known = 1; }
        }
        if ($known == 0) {
            # add to list
            
            (my $class = $res) =~ s/.*class: 0x([a-zA-Z0-9]*).*/$1/;
            (my $time  = $res) =~ s/.*time: ([0-9\.]* [0-9:]*).*/$1/;
            @filtered_results = (@filtered_results, "ind(inquiry_result_event, $ind_msgnr, $addr, 0x$class, rssi, \"$time\", $lat, $long)");
            $ind_msgnr++;
        }
    }

    foreach my $fres (@filtered_results) {
        (my $faddr = $fres) =~ s/.*inquiry_result_event,[^,]*, ([^,]*),.*/$1/;
        $faddr = uc($faddr);

        if (defined($rssi_cnt{$faddr})) {
            $rssi_val{$faddr} = sprintf("%x", floor($rssi_val{$faddr}/$rssi_cnt{$faddr}));
            $rssi_val{$faddr} =~ s/.*(....)/0x$1/;
            dprint "rssi for $faddr is $rssi_val{$faddr}\n";
            $fres =~ s/rssi/$rssi_val{$faddr}/;
        }
        else {
            $fres =~ s/rssi/0x0/;
        }
    }

    return @filtered_results;
}



#------------------------------------------------------------------------------
#   main starts here
#------------------------------------------------------------------------------

$user     = shift || 'hugo';
$pass     = shift || 'oguh';
$remote   = shift || 'localhost';
$port     = shift || 4444; 
$lat      = shift || 0;
$long     = shift || 0;
$has_rssi = shift || 0;
my ($v_mode, $sock);

dprint "my pid is $$\n";
$IO::Socket::SSL::DEBUG = 2;

if ($port =~ /\D/) { $port = getservbyname($port, 'tcp') }
die "No port" unless $port;
$iaddr   = inet_aton($remote)               || die "no host: $remote";

# SSL
#
# needed files: (SSL_use_cert == 1)
#   certs/my-ca.pem
#   certs/client-key.pem
#   certs/client-cert.pem
#
# SSL_verify_mode:
#   0x00: no authentication
#   0x01: verify peer
#   0x02: 0x02 (fail verification if no peer certificate exists; ignored for clients)
#   0x04 (verify client once)
#
# SSL_version: (case-insensitive)
#   'SSLv2/3': (default) auto-negotiates between SSLv2 and SSLv3
#   'SSLv2':
#   'SSLv3':
#   'TLSv1':

dprint "connect to $remote($iaddr) on port $port\n";

if(!($sock = IO::Socket::SSL->new( PeerAddr => "$remote",
                                   PeerPort => "$port",
                                   Proto    => 'tcp',
#                                   SSL_use_cert => 1,
                                   SSL_verify_mode => 0x00,
                                   SSL_version => 'SSLv2',
                                 ))) {
  print STDERR "unable to create socket: '$!'.\n";
  exit(0);
} else {
  print STDERR "connect ($sock).\n" if ($IO::Socket::SSL::DEBUG);
}

# check server cert.
my ($peer_cert, $subject_name, $issuer_name, $cipher);
if( ref($sock) eq "IO::Socket::SSL") {
    if(($peer_cert = $sock->get_peer_certificate)) {
        $subject_name = $peer_cert->subject_name;
        $issuer_name = $peer_cert->issuer_name;
        $cipher = $sock->get_cipher();
    }
    print STDERR "cipher: $cipher.\n";
    print STDERR "server cert:\n".
        "\t '$subject_name' \n\t '$issuer_name'.\n\n";
}
                                                                                

#------------------------------------------------------------------------------
#   To have perl not buffering its communication to the server, set the
#   autoflush option.
#------------------------------------------------------------------------------
$sock->autoflush(1);
my $parent = $$;

my $child_pid = fork;
defined $child_pid or die "fork: $!";

if ($child_pid == 0) {
    #------------------------------------------------------------------------------
    #   This is the process reading and processing commands from the BLT server
    #------------------------------------------------------------------------------
    print "socket reader is running...\n";
    while (defined($line = <$sock>)) {
        dprint $line;
        $line =~ tr/\r\n//d;
        my @request = split(/[\(,\)]/, $line);

        if ($request[0] eq "req")
        {
            #------------------------------------------------------------------------------
            #   Yep, we got a request from the BLT server
            #------------------------------------------------------------------------------

            dprint "message nr is $request[2]\n";
            dprint "server command is $request[1]\n";

            if ($request[1] eq "inquiry") {
                #------------------------------------------------------------------------------
                #   Starting the inquiry procedure
                #------------------------------------------------------------------------------
                my $len;
                my $num_rsp;

                $num_rsp = evalnumber($request[3]);
                $len = evalnumber($request[4]);

                print "start inquiry with len=$len and num_rsp=$num_rsp\n";
                `./inquiry $len $num_rsp`;
            }
            elsif ($request[1] eq "configuration") {
                #------------------------------------------------------------------------------
                #   Configuring our Bluetooth device
                #------------------------------------------------------------------------------
                my $iscan;
                my $pscan;
                my $ptmo;

                $iscan = evalnumber($request[3]);
                $pscan = evalnumber($request[4]);
                $ptmo = evalnumber($request[5]);
                dprint "name is $request[6]\n";
                (my $name = $request[6]) =~ s/[^\"]*\"(.*)\"/$1/;

                print "configuration with iscan=$iscan and pscan=$pscan and ptmo=$ptmo and name '$name'\n";
                $ptmo = $ptmo * 2048;
                if (length($name) > 0) { $name = "name \"" . $name . "\""; } else { $name = ""; }

                #------------------------------------------------------------------------------
                #   NOTE: We are a bit lazy here and switch on/off BOTH page and inquiry scan
                #   by only evaluating the inquiry scan parameter.
                #------------------------------------------------------------------------------
                if ($iscan == 0) {
                    `hciconfig hci0 noscan pageto $ptmo $name`;
                } else {
                    `hciconfig hci0 piscan pageto $ptmo $name`;
                }
            }
            elsif ($request[1] eq "name_request") {
                #------------------------------------------------------------------------------
                #   Starting a Remote Name Request
                #------------------------------------------------------------------------------
                print "start name_req for $request[3]\n";
                `./name_req $request[3]`;
            }
            elsif ($request[1] eq "authentication") {
                #------------------------------------------------------------------------------
                #   Got an Authentication Request
                #------------------------------------------------------------------------------
                print "answering authentication request\n";
                print $sock "ind(authentication, $request[2], \"$user\", \"$pass\")\r\n";
            }
        }
    }
    print "socket reader stopped.\n";
    `killall hcisniff`;
    close($sock);
    exit;
}


#------------------------------------------------------------------------------
#   This is the other process listening for different HCI events.
#   A seperate tool 'hcisniff' is used for this purpose. To have
#   'hcisniff' report RSSI values of Inquiry Responses, the variable
#   $has_rssi has to be set to "has_rssi" on the command line.
#------------------------------------------------------------------------------

$SIG{HUP} = \&catch_zap;

if ($has_rssi eq "has_rssi") {
    if (!open(EVENTS, "./hcisniff has_rssi |")) {
        my $err = "hcisniff could not be opened: $!";
        kill 9,$child_pid; waitpid($child_pid,0);
        die $err;
    }
} else {
    if (!open(EVENTS, "./hcisniff|")) {
        my $err = "hcisniff could not be opened: $!";
        kill 9,$child_pid; waitpid($child_pid,0);
        die $err;
    }
}

print "socket writer is running... \n";

while (defined(my $event = <EVENTS>)) {

    dprint "event: $event";
    $event =~ tr/\r\n//d;
    (my $eventcode = $event) =~ s/^([^ :]*).*/$1/;

    if ($eventcode eq "inquiry_result") {
        #------------------------------------------------------------------------------
        #   hcisniff got an Inquiry Result Event
        #------------------------------------------------------------------------------
        print "got inquiry result\n";
        my @ltime = localtime(time());
        $ltime[5] += 1900;
        $event = sprintf("%s, time: %02d.%02d.%04d %02d:%02d:%02d", $event, $ltime[4]+1, $ltime[3], $ltime[5], $ltime[2], $ltime[1], $ltime[0]);
        @inq_results = (@inq_results, $event);
    }
    elsif ($eventcode eq "inquiry_complete") {
        #------------------------------------------------------------------------------
        #   hcisniff got an Inquiry Complete Event
        #------------------------------------------------------------------------------
        dprint "got inquiry complete\n";
        @inq_results = filter_inq_results(@inq_results);
        foreach my $res (@inq_results) {
            print "$res\n";
            print $sock "$res\r\n";
        }
        @inq_results = ();
        print "ind(inquiry_complete, $ind_msgnr)\n";
        print $sock "ind(inquiry_complete, $ind_msgnr)\r\n";
        $ind_msgnr++;
    }
    elsif ($eventcode eq "name_complete") {
        #------------------------------------------------------------------------------
        #   hcisniff got a Name Request Complete Event
        #------------------------------------------------------------------------------
        dprint "got name complete\n";

        (my $addr = $event) =~ s/.*bdaddr ([0-9a-zA-Z:]*).*/$1/;
        (my $name = $event) =~ s/[^\"]*\"(.*)\".*/$1/;
        # escape " and \
        $name =~ s/([\"\\])/\\$1/g;

        print "ind(name_request, $ind_msgnr, $addr, \"$name\")\n";
        print $sock "ind(name_request, $ind_msgnr, $addr, \"$name\")\r\n";
        $ind_msgnr++;
    }
}

print "exiting socket writer\n";

if ($child_pid != 0) { kill 9,$child_pid; waitpid($child_pid,0); close(EVENTS); }

close ($sock)            || die "close: $!";
exit;

